# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.views.decorators.csrf import csrf_exempt
from django.template.response import TemplateResponse
from blti.config import get_tool_conf, get_launch_data_storage
from blti.oidc_login import BLTIOIDCLogin
import logging


logger = logging.getLogger(__name__)


@csrf_exempt
def login(request):
    try:
        tool_conf = get_tool_conf()
        launch_data_storage = get_launch_data_storage()
        oidc_login = BLTIOIDCLogin(
            request, tool_conf, launch_data_storage=launch_data_storage)
        target_link_uri = getattr(request, request.method)['target_link_uri']
        js_redirect = request.GET.get('lti_storage_frame', None) is not None

        if target_link_uri.startswith('http:') and request.is_secure():
            target_link_uri = f"https:{target_link_uri[5:]}"

        return oidc_login.enable_check_cookies().redirect(
            target_link_uri, js_redirect)
    except KeyError:
        logger.error(f"Missing 'target_link_uri' in {request.method} params: "
                     "{request.body.decode('utf-8')}")
        return TemplateResponse(request, 'blti/500.html',
                                context={'error': 'Missing Target Link URI'},
                                status=500)
    except Exception as ex:
        logger.exception(f"LTI 1.3 login exception: {ex}")
        return TemplateResponse(request, 'blti/500.html',
                                context={'error': str(ex)}, status=500)



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.http import HttpResponse
from pylti1p3.contrib.django.redirect import DjangoRedirect
import json


class BLTILaunchRedirect(DjangoRedirect):
    def __init__(self, location, params, auth_origin):
        self._location = location
        self._params = params
        self._auth_origin = auth_origin
        super().__init__(location)

    def do_js_redirect(self):
        return self._process_response(
            HttpResponse(
                f"""\
                <script type="text/javascript">
                const redirect_location = "{self._location}",
                      parameters = {json.dumps(self._params, indent=4)},
                      redirect_origin = "{self._auth_origin}";
                """
                """\
                var client_data = {
                        nonce: null,
                        state: null,
                        session_cookie_value: null
                    };

                function doRedirection() {
                    var f = document.createElement('form');
                    f.action = redirect_location;
                    f.method = 'POST';

                    formInput(f, 'lti1p3_session_id',
                        client_data.session_cookie_value);
                    formInput(f, 'lti1p3_state',
                        client_data.state);
                    formInput(f, 'lti1p3_nonce',
                        client_data.nonce);

                    for (const p in parameters) {
                        formInput(f, p, parameters[p]);
                    }

                    document.body.appendChild(f);
                    f.submit();
                }

                function formInput(f, k, v) {
                    var i=document.createElement('input');
                    i.type='hidden';
                    i.name=k;
                    i.value=v;
                    f.appendChild(i);
                }

                function clientDataMessageId(prop) {
                    return prop + '_' + parameters.state;
                }

                function getClientData(frame) {
                    for (const prop in client_data) {
                        ltiClientStore(frame, {
                            subject: 'lti.get_data',
                            message_id: clientDataMessageId(prop),
                            key: prop
                        });
                    }
                }

                function dataFetched(frame) {
                    for (const prop in client_data) {
                        if (!client_data[prop]) {
                            return false;
                        }
                    }
                    return true;
                }

                function ltiClientStore(frame, data) {
                    window.parent.frames[frame].postMessage(
                        data, redirect_origin);
                }

                function ltiClientStoreResponse(event) {
                    var message = event.data;
                    switch (message.subject) {
                        case 'lti.capabilities.response':
                            var supported = message.supported_messages;
                            for (var i = 0; i < supported.length; i++) {
                                if (supported[i].subject == "lti.get_data") {
                                    getClientData(supported[i].frame);
                                }
                            }
                        break;
                        case 'lti.get_data.response':
                            client_data[message.key] = message.value;
                            if (dataFetched()) {
                                doRedirection();
                            }
                        break;
                    }

                    if (message.error) {
                        console.error("event " + message.subject +
                                      " error (" + message.error.code +
                                      "): " + message.error.message);
                    }
                }

                function clientStoreAndRedirect() {
                    window.parent.postMessage({
                        subject: 'lti.capabilities'}, '*');
                }

                window.addEventListener("message", ltiClientStoreResponse);
                document.addEventListener("DOMContentLoaded",
                                          clientStoreAndRedirect);
                </script>
                """
            )
        )



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from .base import BLTIView
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from blti.config import get_tool_conf, get_launch_data_storage
from blti.request import BLTIRequest
from blti.exceptions import BLTIException
from blti.validators import BLTIRequestValidator
from blti.launch_redirect import BLTILaunchRedirect
from blti.cookie import BLTICookieService
from pylti1p3.exception import OIDCException
from pylti1p3.contrib.django import DjangoMessageLaunch
from oauthlib.oauth1.rfc5849.endpoints.signature_only import (
    SignatureOnlyEndpoint)
from pylti1p3.contrib.django.session import DjangoSessionService
from urllib.parse import urljoin, urlparse, urlencode
import json
import logging


logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name='dispatch')
class BLTILaunchView(BLTIView):
    http_method_names = ['get', 'post']
    _oidc_launch_keys = ['state', 'authenticity_token', 'id_token', 'utf8']
    _oidc_launch_client_store_state = 'lti1p3_state'
    _oidc_launch_client_store_session_id = 'lti1p3_session_id'
    _oidc_launch_client_store_nonce = 'lti1p3_nonce'

    def post(self, request, *args, **kwargs):
        return self._launch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        return self._launch(request, *args, **kwargs)

    def _launch(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def dispatch(self, request, *args, **kwargs):
        try:
            launch_data = self.validate_1p1(request)
            logger.debug(f"LTI 1.1 launch")
        except BLTIException as ex:
            try:
                if self._missing_lti_parameters(request):
                    return self._client_store_redirect(request)

                launch_data = self.validate_1p3(request)
                logger.debug(f"LTI 1.3 launch")
            except OIDCException as ex:
                logger.error(f"LTI authentication failure: {ex}")
                self.template_name = 'blti/401.html'
                return self.render_to_response(
                    {'LTI authentication failure': str(ex)}, status=401)
            except Exception as ex:
                logger.error(f"LTI launch error: {ex}")
                self.template_name = 'blti/401.html'
                return self.render_to_response(
                    {'LTI launch failure': str(ex)}, status=401)

        self.set_session(**launch_data)
        return super(BLTILaunchView, self).dispatch(request, *args, **kwargs)

    def validate_1p1(self, request):
        request_validator = BLTIRequestValidator()
        endpoint = SignatureOnlyEndpoint(request_validator)
        uri = request.build_absolute_uri()
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        valid, oauth_req = endpoint.validate_request(
            uri, request.method, request.body, headers)

        # if non-ssl fixup scheme to validate signature generated
        # on the other side of ingress
        if not valid and uri.startswith('http:') and request.is_secure():
            valid, oauth_req = endpoint.validate_request(
                f"https{uri[4:]}", request.method, request.body, headers)

        if not valid:
            raise BLTIException('Invalid OAuth Request')

        blti_params = dict(oauth_req.params)
        return blti_params

    def validate_1p3(self, request):
        tool_conf = get_tool_conf()
        launch_data_storage = get_launch_data_storage()
        message_launch = DjangoMessageLaunch(
            request, tool_conf, launch_data_storage=launch_data_storage)
        return message_launch.get_launch_data()

    def _missing_lti_parameters(self, request):
        blti_request = BLTIRequest(request)
        cookie_service = BLTICookieService(blti_request)
        session_service = DjangoSessionService(request)
        data_storage = get_launch_data_storage()
        data_storage.set_request(blti_request)

        session_cookie_name = data_storage.get_session_cookie_name()
        session_id = cookie_service.get_cookie(session_cookie_name)
        if not session_id:
            # peel parameters inserted from client side storage
            # off and insert them into the request validation
            session_id = self.get_parameter(
                request, self._oidc_launch_client_store_session_id)
            if session_id:
                # insert request session cookie
                cookie_service.set_request_cookie(
                    session_cookie_name, session_id)

                # insert request state cookie
                state = self.get_parameter(
                    request, self._oidc_launch_client_store_state)
                cookie_service.set_request_cookie(state, state)

                # add nonce to session
                nonce = self.get_parameter(
                    request, self._oidc_launch_client_store_nonce)
                session_service.save_nonce(nonce)

                # fall thru to 1.3 launch

            elif self.get_parameter(request, 'lti_storage_target'):
                return True

        return False

    def _client_store_redirect(self, request):
        redirect_uri = request.build_absolute_uri()

        if redirect_uri.startswith('http:') and request.is_secure():
            redirect_uri = f"https{uri[4:]}"

        parameters = {k: self.get_parameter(request, k) for (
            k) in self._oidc_launch_keys}
        auth_origin = self._login_origin_from_iss(parameters['id_token'])

        logger.debug(f"LTI 1.3 client side storage redirect")
        return BLTILaunchRedirect(
            redirect_uri, parameters, auth_origin).do_js_redirect()

    def _login_origin_from_iss(self, id_token):
        # oidc auth origin from iss dug out of jwt
        jwt = self._jwt_from_id_token(id_token)
        iss = jwt.get('iss')
        reg = get_tool_conf().find_registration_by_issuer(iss)
        auth_url = urlparse(reg.get_auth_token_url())
        return f"{auth_url.scheme}://{auth_url.netloc}"

    def _jwt_from_id_token(self, id_token):
        parts = id_token.split('.')
        if len(parts) != 3:
            raise BLTIException("Invalid id_token")

        raw_body = DjangoMessageLaunch.urlsafe_b64decode(parts[1])
        return json.loads(raw_body)

    def get_parameter(self, request, key):
        params = request.POST if request.method == 'POST' else request.GET
        return params.get(key)



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from blti.redirect import BLTIRedirect
from blti.cookies_allowed_check import BLTICookiesAllowedCheckPage
from pylti1p3.request import Request
from pylti1p3.contrib.django import DjangoOIDCLogin
from pylti1p3.contrib.django.request import DjangoRequest
from pylti1p3.contrib.django.cookie import DjangoCookieService


class BLTIOIDCLogin(DjangoOIDCLogin):
    def __init__(
        self,
        request,
        tool_config,
        session_service=None,
        cookie_service=None,
        launch_data_storage=None,
    ):
        django_request = (
            request if isinstance(
                request, Request) else DjangoRequest(request)
        )
        cookie_service = (
            cookie_service if (
                cookie_service) else DjangoCookieService(django_request)
        )
        super().__init__(
            django_request,
            tool_config,
            session_service,
            cookie_service,
            launch_data_storage,
        )

    def get_cookies_allowed_js_check(self) -> str:
        protocol = "https" if self._request.is_secure() else "http"
        params_lst = [
            "iss",
            "login_hint",
            "target_link_uri",
            "lti_message_hint",
            "lti_deployment_id",
            "client_id",
            "lti_storage_target"
        ]
        additional_login_params = self.get_additional_login_params()
        params_lst.extend(additional_login_params)

        params = {"lti1p3_new_window": "1"}
        for param_key in params_lst:
            param_value = self._get_request_param(param_key)
            if param_value:
                params[param_key] = param_value

        page = BLTICookiesAllowedCheckPage(
            params,
            protocol,
            self._cookies_unavailable_msg_main_text,
            self._cookies_unavailable_msg_click_text,
            self._cookies_check_loading_text,
        )

        return page.get_html()

    def get_redirect(self, url):
        return BLTIRedirect(
            url, cookie_service=self._cookie_service,
            session_service=self._session_service)



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from pylti1p3.cookies_allowed_check import CookiesAllowedCheckPage
from html import escape
import json


class BLTICookiesAllowedCheckPage(CookiesAllowedCheckPage):
    def get_js_block(self) -> str:
        js_block = """\
        var siteProtocol = '%s';
        var urlParams = %s;
        var htmlEntities = {
            "&lt;": "<",
            "&gt;": ">",
            "&amp;": "&",
            "&quot;": '"',
            "&#x27;": "'"
        };

        function unescapeHtmlEntities(str) {
            for (var htmlCode in htmlEntities) {
                str = str.replace(
                    new RegExp(htmlCode, "g"), htmlEntities[htmlCode]);
            }
            return str;
        }

        function getUpdatedUrl(lti_storage_frame) {
            var newSearchParams = [];
            for (var key in urlParams) {
                if (window.location.search.indexOf(key + '=') === -1) {
                    newSearchParams.push(key + '=' + encodeURIComponent(
                        unescapeHtmlEntities(urlParams[key])));
                }
            }
            if (lti_storage_frame) {
                newSearchParams.push('lti_storage_frame=' +
                                     encodeURIComponent(lti_storage_frame));
            }
            var searchParamsStr = newSearchParams.join('&');
            if (window.location.search !== '') {
                searchParamsStr = window.location.search +
                    '&' + searchParamsStr;
            } else {
                searchParamsStr = '?' + searchParamsStr;
            }
            return window.location.protocol + '//' + window.location.hostname +
                (window.location.port ? (":" + window.location.port) : "") +
                window.location.pathname + searchParamsStr;
        }

        function displayLoadingBlock() {
            document.getElementById(
                "lti1p3-loading-msg").style.display = "block";
        }

        function displayWarningBlock() {
            document.getElementById(
                "lti1p3-warning-msg").style.display = "block";
            var newTabLink = document.getElementById("lti1p3-new-tab-link");
            var contentUrl = getUpdatedUrl();
            newTabLink.onclick = function() {
                window.open(contentUrl , '_blank');
                newTabLink.parentNode.removeChild(newTabLink);
            };
        }

        function checkCookiesAllowed() {
            var cookie = "lti1p3_test_cookie=1; path=/";
            if (siteProtocol === 'https') {
                cookie = cookie + '; Partitioned; SameSite=None; Secure';
            }

            document.cookie = cookie;
            var res = document.cookie.indexOf("lti1p3_test_cookie") !== -1;
            if (res) {
                // remove test cookie and reload page
                document.cookie = "lti1p3_test_cookie=1; " +
                                  "expires=Thu, 01-Jan-1970 00:00:01 GMT";
                displayLoadingBlock();
                window.location.href = getUpdatedUrl();
            } else if ('lti_storage_target' in urlParams) {
                var frame = urlParams.lti_storage_target;

                displayLoadingBlock();
                window.location.href = getUpdatedUrl(frame);
            } else {
                displayWarningBlock();
            }
        }

        document.addEventListener("DOMContentLoaded", checkCookiesAllowed);
        """
        # pylint: disable=deprecated-method
        js_block = js_block % (
            self._protocol,
            json.dumps({k: escape(v, True) for k, v in self._params.items()}),
        )
        return js_block



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from pylti1p3.contrib.django.request import DjangoRequest


class BLTIRequest(DjangoRequest):
    def __init__(self, request, post_only=False, default_params=None):
        super().__init__(request, post_only, default_params)

    def set_request_cookie(self, key, value):
        self._request.COOKIES[key] = value



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from pylti1p3.contrib.django.cookie import DjangoCookieService


class BLTICookieService(DjangoCookieService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_request_cookie(self, name, value):
        self._request.set_request_cookie(self._get_key(name), value)



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


"""
Cross Site Request Forgery and Session HTTP Header Middleware.

This module provides middleware that implements protection
against request forgeries from other sites.
"""

from django.conf import settings
from django.contrib.auth import authenticate, login
from blti import BLTI
from blti.exceptions import BLTIException


class CSRFHeaderMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        csrf_token = request.META.get('HTTP_X_CSRFTOKEN', None)
        if csrf_token is not None:
            csrf_token_name = settings.CSRF_COOKIE_NAME
            request.COOKIES[csrf_token_name] = csrf_token

        return self.get_response(request)


class SessionHeaderMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        session_id = request.META.get('HTTP_X_SESSIONID', None)
        if session_id is not None:
            session_key = settings.SESSION_COOKIE_NAME
            request.COOKIES[session_key] = session_id

        return self.get_response(request)


class LTISessionAuthenticationMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        try:
            lti_launch_parameters = BLTI().get_session(request)
            lti_user = lti_launch_parameters.get(
                "https://purl.imsglobal.org/spec/lti/claim/custom", {}).get(
                    "canvas_user_login_id")
            if lti_user:
                user = authenticate(request, remote_user=lti_user)
                login(request, user)
        except BLTIException:
            pass

        return self.get_response(request)


class SameSiteMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        for cookie in ['sessionid', 'csrftoken']:
            if cookie in response.cookies:
                response.cookies[cookie]['samesite'] = 'None'

        return response



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.http import HttpResponse  # type: ignore
from pylti1p3.contrib.django.redirect import DjangoRedirect
from urllib.parse import urlparse, parse_qs
from uuid import uuid4


class BLTIRedirect(DjangoRedirect):
    def __init__(self, location, cookie_service=None, session_service=None):
        self._session_cookie_value = (
            f"{session_service.data_storage.get_session_id()}")
        self._location = location
        super().__init__(location, cookie_service)

    def do_js_redirect(self):
        return self._process_response(
            HttpResponse(
                f"""\
                <script type="text/javascript">
                const redirect_location = "{self._location}",
                      parsed_location = URL.parse(redirect_location),
                      parsed_params = parsed_location.searchParams,
                      origin = parsed_location.origin,
                      nonce = parsed_params.get('nonce'),
                      state = parsed_params.get('state'),
                      session_cookie_value = "{self._session_cookie_value}";
                """
                """
                var client_data = {
                        nonce: {
                            value: nonce,
                            stored: false
                        },
                        state: {
                            value: state,
                            stored: false
                        },
                        session_cookie_value: {
                            value: session_cookie_value,
                            stored: false
                        }
                    };

                function doRedirection() {
                    window.location=redirect_location;
                }

                function validClientData() {
                    for (const prop in client_data) {
                        if (!client_data[prop].value) {
                            return false;
                        }
                    }

                    return true;
                }

                function clientDataMessageId(prop) {
                    return prop + '_' + state;
                }

                function putClientData(frame) {
                    for (const prop in client_data) {
                        ltiClientStore(frame, {
                            subject: 'lti.put_data',
                            message_id: clientDataMessageId(prop),
                            key: prop,
                            value:  client_data[prop].value
                        });
                    }
                }

                function dataStored(frame) {
                    for (const prop in client_data) {
                        if (!client_data[prop].stored) {
                            return false;
                        }
                    }
                    return true;
                }

                function ltiClientStore(frame, data) {
                    window.parent.frames[frame].postMessage(data, origin);
                }

                function ltiClientStoreResponse(event) {
                    var message = event.data;
                    switch (message.subject) {
                        case 'lti.capabilities.response':
                            var supported = message.supported_messages;
                            for (var i = 0; i < supported.length; i++) {
                                if (supported[i].subject == "lti.put_data") {
                                    putClientData(supported[i].frame);
                                }
                            }
                        break;
                        case 'lti.put_data.response':
                            client_data[message.key].stored = true;
                            if (dataStored()) {
                                doRedirection();
                            }
                        break;
                    }

                    if (message.error) {
                        console.error("event error:" + message.subject +
                                      ", code: " + message.error.code +
                                      ", message: " + message.error.message);
                    }
                }

                function clientStoreAndRedirect() {
                    if (validClientData()) {
                        window.parent.postMessage(
                             {subject: 'lti.capabilities'}, '*');
                        setTimeout(doRedirection, 5000);
                    } else {
                        doRedirection();
                    }
                }

                window.addEventListener("message", ltiClientStoreResponse);
                document.addEventListener(
                    "DOMContentLoaded", clientStoreAndRedirect);
                </script>
                """
            )
        )



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from django.conf import settings
from django.test import TestCase, Client
from django.urls import reverse
from unittest.mock import patch
from blti.views.launch import BLTILaunchView


class TestLaunchViews(TestCase):
    def setup(self, request, *args, **kwargs):
        self.client = Client()

    def test_dev_launch_view(self):
        response = self.client.get(reverse('dev-launch'), {
            'role': 'Instructor',
            'campus': 'seattle',
        }, secure=True)

        self.assertEqual(response.status_code, 200)

        response = self.client.post(reverse('lti-launch'), secure=True)
        self.assertEqual(response.status_code, 200)

    @patch('blti.views.launch.BLTILaunchView._login_origin_from_iss')
    def test_client_store_redirect(self, mocked):
        mocked.return_value = 'https://example.com'
        response = self.client.post(reverse('lti-launch-data'), {
            'state': 'state-ac00bf57-bdd7-47c8-8b95-918f94797aef',
            'authenticity_token': 'T1o1dHJVeGNKRXhsTnh0eGRLCg==',
            'id_token': 'MmZQdHZZSFcwd084RjAxZ1BlUUkwTVpZWXhDa1Uza2JDMmxSCg==',
            'utf8': '✓',
            'lti_storage_target': 'client_store'
        }, secure=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'getClientData')
        self.assertContains(response, 'ltiClientStoreResponse')
        self.assertContains(response, 'doRedirection')
